@@ -1,7 +1,7 @@
 package DBIx::Custom;
 use Object::Simple -base;
 
-our $VERSION = '0.1735';
+our $VERSION = '0.1736';
 use 5.008001;
 
 use Carp 'croak';
@@ -123,27 +123,28 @@ sub assign_clause {
     my ($self, $param, $opts) = @_;
     
     my $wrap = $opts->{wrap} || {};
-
+    my $safety = $self->{safety_character} || $self->safety_character;
     my $qp = $self->_q('');
     my $q = substr($qp, 0, 1) || '';
     my $p = substr($qp, 1, 1) || '';
     
-    # Create set tag
-    my @params;
-    my $safety = $self->safety_character;
-    for my $column (sort keys %$param) {
-        croak qq{"$column" is not safety column name } . _subname
-          unless $column =~ /^[$safety\.]+$/;
-        my $column_quote = "$q$column$p";
-        $column_quote =~ s/\./$p.$q/;
-        my $func = $wrap->{$column} || sub { $_[0] };
-        push @params,
-          ref $param->{$column} eq 'SCALAR' ? "$column_quote = " . ${$param->{$column}}
-        : "$column_quote = " . $func->(":$column");
+    # Check unsafety keys
+    unless ((join('', keys %$param) || '') =~ /^[$safety\.]+$/) {
+        for my $column (keys %$param) {
+            croak qq{"$column" is not safety column name } . _subname
+              unless $column =~ /^[$safety\.]+$/;
+        }
     }
-    my $tag = join(', ', @params);
     
-    return $tag;
+    # Assign clause (performance is important)
+    join(
+      ', ',
+      map {
+          ref $param->{$_} eq 'SCALAR' ? "$q$_$p = " . ${$param->{$_}}
+          : $wrap->{$_} ? "$q$_$p = " . $wrap->{$_}->(":$_")
+          : "$q$_$p = :$_";
+      } sort keys %$param
+    );
 }
 
 sub column {
@@ -254,7 +255,7 @@ sub delete {
     $sql .= "from " . $self->_q($opt{table}) . " $w->{clause} ";
     
     # Execute query
-    return $self->execute($sql, $w->{param}, %opt);
+    $self->execute($sql, $w->{param}, %opt);
 }
 
 sub delete_all { shift->delete(allow_delete_all => 1, @_) }
@@ -361,7 +362,8 @@ sub execute {
     $param ||= $opt{param} || {};
     my $tables = $opt{table} || [];
     $tables = [$tables] unless ref $tables eq 'ARRAY';
-    my $filter = _array_to_hash($opt{filter});
+    my $filter = ref $opt{filter} eq 'ARRAY' ?
+      _array_to_hash($opt{filter}) : $opt{filter};
     
     # Append
     $sql .= $opt{append} if defined $opt{append} && !ref $sql;
@@ -377,7 +379,7 @@ sub execute {
     }
         
     # Save query
-    $self->last_sql($query->sql);
+    $self->{last_sql} = $query->{sql};
 
     # Return query
     return $query if $opt{query};
@@ -401,24 +403,26 @@ sub execute {
     
     # Type rule
     my $type_filters = {};
-    unless ($opt{type_rule_off}) {
-        my $type_rule_off_parts = {
-            1 => $opt{type_rule1_off},
-            2 => $opt{type_rule2_off}
-        };
-        for my $i (1, 2) {
-            unless ($type_rule_off_parts->{$i}) {
-                $type_filters->{$i} = {};
-                my $table_alias = $opt{table_alias} || {};
-                for my $alias (keys %$table_alias) {
-                    my $table = $table_alias->{$alias};
-                    
-                    for my $column (keys %{$self->{"_into$i"}{key}{$table} || {}}) {
-                        $type_filters->{$i}->{"$alias.$column"} = $self->{"_into$i"}{key}{$table}{$column};
+    if ($self->{_type_rule_is_called}) {
+        unless ($opt{type_rule_off}) {
+            my $type_rule_off_parts = {
+                1 => $opt{type_rule1_off},
+                2 => $opt{type_rule2_off}
+            };
+            for my $i (1, 2) {
+                unless ($type_rule_off_parts->{$i}) {
+                    $type_filters->{$i} = {};
+                    my $table_alias = $opt{table_alias} || {};
+                    for my $alias (keys %$table_alias) {
+                        my $table = $table_alias->{$alias};
+                        
+                        for my $column (keys %{$self->{"_into$i"}{key}{$table} || {}}) {
+                            $type_filters->{$i}->{"$alias.$column"} = $self->{"_into$i"}{key}{$table}{$column};
+                        }
                     }
+                    $type_filters->{$i} = {%{$type_filters->{$i}}, %{$self->{"_into$i"}{key}{$main_table} || {}}}
+                      if $main_table;
                 }
-                $type_filters->{$i} = {%{$type_filters->{$i}}, %{$self->{"_into$i"}{key}{$main_table} || {}}}
-                  if $main_table;
             }
         }
     }
@@ -449,17 +453,22 @@ sub execute {
     }
     
     # Create bind values
-    my $bind = $self->_create_bind_values($param, $query->columns,
-      $filter, $type_filters, _array_to_hash($opt{bind_type} || $opt{type}));
+    my ($bind, $bind_types) = $self->_create_bind_values($param, $query->columns,
+      $filter, $type_filters, $opt{bind_type} || $opt{type} || {});
 
     # Execute
-    my $sth = $query->sth;
+    my $sth = $query->{sth};
     my $affected;
     eval {
-        $sth->bind_param($_ + 1, $bind->[$_]->{value},
-            $bind->[$_]->{bind_type} ? $bind->[$_]->{bind_type} : ())
-          for (0 .. @$bind - 1);
-        $affected = $sth->execute;
+        if ($opt{bind_type} || $opt{type}) {
+            $sth->bind_param($_ + 1, $bind->[$_],
+                $bind_types->[$_] ? $bind_types->[$_] : ())
+              for (0 .. @$bind - 1);
+            $affected = $sth->execute;
+        }
+        else {
+            $affected = $sth->execute(@$bind);
+        }
     };
     
     $self->_croak($@, qq{. Following SQL is executed.\n}
@@ -479,39 +488,36 @@ sub execute {
         print STDERR "Bind values: " . join(', ', @output) . "\n\n";
     }
     
-    # Select statement
-    if ($sth->{NUM_OF_FIELDS}) {
-        
-        # Filter(DEPRECATED!)
-        my $filter = {};
-        if ($self->{filter}{on}) {
-            $filter->{in}  = {};
-            $filter->{end} = {};
-            push @$tables, $main_table if $main_table;
-            for my $table (@$tables) {
-                for my $way (qw/in end/) {
-                    $filter->{$way} = {%{$filter->{$way}},
-                      %{$self->{filter}{$way}{$table} || {}}};
-                }
+    # Not select statement
+    return $affected unless $sth->{NUM_OF_FIELDS};
+
+    # Filter(DEPRECATED!)
+    my $infilter = {};
+    if ($self->{filter}{on}) {
+        $infilter->{in}  = {};
+        $infilter->{end} = {};
+        push @$tables, $main_table if $main_table;
+        for my $table (@$tables) {
+            for my $way (qw/in end/) {
+                $infilter->{$way} = {%{$infilter->{$way}},
+                  %{$self->{filter}{$way}{$table} || {}}};
             }
         }
-        
-        # Result
-        my $result = $self->result_class->new(
-            sth => $sth,
-            dbi => $self,
-            default_filter => $self->{default_in_filter},
-            filter => $filter->{in} || {},
-            end_filter => $filter->{end} || {},
-            type_rule => {
-                from1 => $self->type_rule->{from1},
-                from2 => $self->type_rule->{from2}
-            },
-        );
-        return $result;
     }
-    # Not select statement
-    else { return $affected }
+    
+    # Result
+    my $result = $self->result_class->new(
+        sth => $sth,
+        dbi => $self,
+        default_filter => $self->{default_in_filter},
+        filter => $infilter->{in} || {},
+        end_filter => $infilter->{end} || {},
+        type_rule => {
+            from1 => $self->type_rule->{from1},
+            from2 => $self->type_rule->{from2}
+        },
+    );
+    $result;
 }
 
 sub get_table_info {
@@ -562,7 +568,7 @@ sub insert {
     # Options
     my $param = @_ % 2 ? shift : undef;
     my %opt = @_;
-    warn "insert method param option is DEPRECATED" if $opt{param};
+    warn "insert method param option is DEPRECATED!" if $opt{param};
     $param ||= delete $opt{param} || {};
     
     # Timestamp
@@ -586,7 +592,7 @@ sub insert {
       . $self->values_clause($param, {wrap => $opt{wrap}}) . " ";
     
     # Execute query
-    return $self->execute($sql, $param, %opt);
+    $self->execute($sql, $param, %opt);
 }
 
 sub insert_timestamp {
@@ -856,7 +862,7 @@ sub select {
     # Execute query
     my $result = $self->execute($sql, $w->{param}, %opt);
     
-    return $result;
+    $result;
 }
 
 sub setup_model {
@@ -918,6 +924,8 @@ sub show_tables {
 
 sub type_rule {
     my $self = shift;
+
+    $self->{_type_rule_is_called} = 1;
     
     if (@_) {
         my $type_rule = ref $_[0] eq 'HASH' ? $_[0] : {@_};
@@ -1023,7 +1031,7 @@ sub update {
     $sql .= $self->_q($opt{table}) . " set $assign_clause $w->{clause} ";
     
     # Execute query
-    return $self->execute($sql, $param, %opt);
+    $self->execute($sql, $param, %opt);
 }
 
 sub update_all { shift->update(allow_update_all => 1, @_) };
@@ -1081,28 +1089,35 @@ sub values_clause {
     my $wrap = $opts->{wrap} || {};
     
     # Create insert parameter tag
-    my $safety = $self->safety_character;
-    my @columns;
-    my @placeholders;
+    my $safety = $self->{safety_character} || $self->safety_character;
     my $qp = $self->_q('');
     my $q = substr($qp, 0, 1) || '';
     my $p = substr($qp, 1, 1) || '';
     
-    for my $column (sort keys %$param) {
-        croak qq{"$column" is not safety column name } . _subname
-          unless $column =~ /^[$safety\.]+$/;
-        my $column_quote = "$q$column$p";
-        $column_quote =~ s/\./$p.$q/;
-        push @columns, $column_quote;
-        
-        my $func = $wrap->{$column} || sub { $_[0] };
-        push @placeholders,
-          ref $param->{$column} eq 'SCALAR' ? ${$param->{$column}}
-        : $func->(":$column");
+    # Check unsafety keys
+    unless ((join('', keys %$param) || '') =~ /^[$safety\.]+$/) {
+        for my $column (keys %$param) {
+            croak qq{"$column" is not safety column name } . _subname
+              unless $column =~ /^[$safety\.]+$/;
+        }
     }
     
-    return '(' . join(', ', @columns) . ') ' . 'values ' .
-           '(' . join(', ', @placeholders) . ')'
+    # values clause(performance is important)
+    '(' .
+    join(
+      ', ',
+      map { "$q$_$p" } sort keys %$param
+    ) .
+    ') values (' .
+    join(
+      ', ',
+      map {
+          ref $param->{$_} eq 'SCALAR' ? ${$param->{$_}} :
+          $wrap->{$_} ? $wrap->{$_}->(":$_") :
+          ":$_";
+      } sort keys %$param
+    ) .
+    ')'
 }
 
 sub where { DBIx::Custom::Where->new(dbi => shift, @_) }
@@ -1185,52 +1200,57 @@ sub _create_query {
 sub _create_bind_values {
     my ($self, $params, $columns, $filter, $type_filters, $bind_type) = @_;
     
+    $bind_type = _array_to_hash($bind_type) if ref $bind_type eq 'ARRAY';
+    
     # Create bind values
-    my $bind = [];
-    my $count = {};
-    my $not_exists = {};
+    my @bind;
+    my @types;
+    my %count;
+    my %not_exists;
     for my $column (@$columns) {
         
-        # Value
-        my $value;
+        # Bind value
         if(ref $params->{$column} eq 'ARRAY') {
-            my $i = $count->{$column} || 0;
-            $i += $not_exists->{$column} || 0;
+            my $i = $count{$column} || 0;
+            $i += $not_exists{$column} || 0;
             my $found;
             for (my $k = $i; $i < @{$params->{$column}}; $k++) {
                 if (ref $params->{$column}->[$k] eq 'DBIx::Custom::NotExists') {
-                    $not_exists->{$column}++;
+                    $not_exists{$column}++;
                 }
                 else  {
-                    $value = $params->{$column}->[$k];
+                    push @bind, $params->{$column}->[$k];
                     $found = 1;
                     last
                 }
             }
             next unless $found;
         }
-        else { $value = $params->{$column} }
+        else { push @bind, $params->{$column} }
         
         # Filter
-        my $f = $filter->{$column} || $self->{default_out_filter} || '';
-        $value = $f->($value) if $f;
+        if (my $f = $filter->{$column} || $self->{default_out_filter} || '') {
+            $bind[-1] = $f->($bind[-1]);
+        }
         
         # Type rule
-        my $tf1 = $self->{"_into1"}->{dot}->{$column}
-          || $type_filters->{1}->{$column};
-        $value = $tf1->($value) if $tf1;
-        my $tf2 = $self->{"_into2"}->{dot}->{$column}
-          || $type_filters->{2}->{$column};
-        $value = $tf2->($value) if $tf2;
+        if ($self->{_type_rule_is_called}) {
+            my $tf1 = $self->{"_into1"}->{dot}->{$column}
+              || $type_filters->{1}->{$column};
+            $bind[-1] = $tf1->($bind[-1]) if $tf1;
+            my $tf2 = $self->{"_into2"}->{dot}->{$column}
+              || $type_filters->{2}->{$column};
+            $bind[-1] = $tf2->($bind[-1]) if $tf2;
+        }
        
-        # Bind values
-        push @$bind, {value => $value, bind_type => $bind_type->{$column}};
+        # Bind types
+        push @types, $bind_type->{$column};
         
         # Count up 
-        $count->{$column}++;
+        $count{$column}++;
     }
     
-    return $bind;
+    return (\@bind, \@types);
 }
 
 sub _id_to_param {
@@ -1417,7 +1437,11 @@ sub _quote {
 sub _q {
     my ($self, $value, $quotemeta) = @_;
     
-    my $quote = $self->_quote;
+    my $quote = $self->{reserved_word_quote}
+      || $self->{quote} || $self->quote || '';
+    return "$quote$value$quote"
+      if !$quotemeta && ($quote eq '`' || $quote eq '"');
+    
     my $q = substr($quote, 0, 1) || '';
     my $p;
     if (defined $quote && length $quote > 1) {
@@ -1639,7 +1663,7 @@ sub assign_param {
 sub update_param {
     my ($self, $param, $opts) = @_;
     
-    warn "update_param is DEPRECATED! use assing_clause instead.";
+    warn "update_param is DEPRECATED! use assign_clause instead.";
     
     # Create update parameter tag
     my $tag = $self->assign_clause($param, $opts);
